////////////////////////////////////////////////////////////////////// 
// KinshipEmp.cpp 
// (c) 2012-2013 Shuang Feng, Dajiang Liu, Goncalo Abecasis
// 
// This file is distributed as part of the RareMetalWorker source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile RareMetalWorker.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday November 28, 2012
// 

#include "KinshipEmp.h"
#include "MathMatrix.h"
#include "PedigreeAlleleFreq.h"
#include "OutputKin.h"
#include "PreMeta.h"
#include <string>
#include <cstring>
#include <iostream>
#include <stdlib.h>
#include <vector>
#include "math.h"
#include <algorithm>

double KinshipEmp::q = 0.05;
double KinshipEmp::miss = 0.05;

void KinshipEmp::Setup(Pedigree & ped,IntArray * pPheno, IntArray & genotypedSamplePED, FILE * log)
{
    printf("Calculating empirical kinship matrix...\n");
    fprintf(log,"Calculating empirical kinship matrix...\n");

    int N=0;
    int total_n = genotypedSamplePED.Length();

    allPairs.Dimension(total_n, total_n);
    allPairs.Zero();
    Vector X;
    X.Dimension(total_n);

    String chr;
    int pos;

    for(int i=0;i<ped.markerCount;i++)
    {
       StringArray markerName;
       markerName.AddTokens(ped.markerNames[i],":");

       if(markerName.Length()<=1)
       {
	  printf("WARNING: marker name %s in DAT file is not in CHR:POS format and will be skipped from calculating empirical kinsihp.\n",ped.markerNames[i].c_str());
	  fprintf(log,"WARNING: marker name %s in DAT file is not in CHR:POS format and will be skipped from calculating empirical kinship.\n",ped.markerNames[i].c_str());
	  continue;
       }

       chr = markerName[0];
       pos = atoi(markerName[1].c_str());

       if(chr==PreMeta::xLabel && pos>=PreMeta::Xstart && pos<=PreMeta::Xend)
	  continue;

       //X.Clear();
       bool missingStatus=false;
       int alleleCount = CountAlleles(ped, i);
       //if no one is genotyped, mornomorphic, or marker is not biallelic
       if(alleleCount!=2)
	  continue;

       double mean=0.0;
       double variance = 0.0;
       double mean_sqr = 0.0;

       //now fill in the genotypes
       int n_=0,nmiss=0;
       double maf = 0.0;
       //only generate matrix for genotyped individuals
       for(int p=0;p<total_n;p++)
       {
	  if(!ped[genotypedSamplePED[p]].isGenotyped(i))
	  {
	     X[p] = -1.0;
	     nmiss++;
	     if(nmiss>total_n*miss)
	     {
		missingStatus=true;
		break;
	     }
	  }
	  else if(ped[genotypedSamplePED[p]].markers[i].isHeterozygous())
	  {
	     X[p]=1.0;
	     mean += X[p];
	     mean_sqr += X[p]*X[p];
	     n_++;
	  }
	  else if(ped[genotypedSamplePED[p]].markers[i].isHomozygous())
	  {
	     if(ped[genotypedSamplePED[p]].markers[i].hasAllele(2))
	     {
		X[p]=2.0;
		mean += X[p];
		mean_sqr += X[p]*X[p];
		n_++;
	     }
	     else if(ped[genotypedSamplePED[p]].markers[i].hasAllele(1))
	     {
		X[p]=0.0;
		n_++;
	     }
	  }
	  else
	  {
	     X[p]=0.0;
	     n_++;
	  }
       }

       if(missingStatus)
	  continue;

       maf = mean/(2.0*n_);

       if(nmiss>total_n*miss || maf<q || 1.0-maf<q)
	  continue;

       N++;
       mean /= n_;
       variance = mean_sqr/n_ - mean*mean;

       for(int p=0;p<total_n;p++)
       {
	  if(X[p]==-1.0)
	     X[p] =0.0;
	  else
	     X[p] -= mean;
       }

       for(int r=0;r<total_n;r++)
       {
	  for(int c=r;c<total_n;c++)
	  {
	     allPairs[r][c] += X[r]*X[c]/variance;
	  }
       }
    }

if(N==0)
error("ERROR! No qualified marker was included in calculating empirical kinsihp matrix.\n");

printf("%d markers were included to calculate empirical kinship.\n",N);
fprintf(log,"%d markers were included to calculate empirical kinship.\n",N);

    for(int i=0;i<total_n;i++)
       for(int j=i+1;j<total_n;j++)
	  allPairs[j][i] = allPairs[i][j];

    allPairs.Multiply(1.0/N);

    printf("Finished calculating empirical kinship matrix. \n ");
    fprintf(log,"Finished calculating empirical kinship matrix. \n ");
    if(OutputKin::outputKin)
    {
       printf("Saving empirical kinship matrix...\n");
       fprintf(log,"Saving empirical kinship matrix...\n");
       FILE * file = NULL;
       String filename;
       if(PreMeta::outputFile != "")
       {
	  if(PreMeta::outputFile.FastFindLastChar('.')!=-1 || PreMeta::outputFile.FastFindLastChar('/')!=-1)
	     filename.printf("%sEmpirical.Kinship",(const char *) PreMeta::outputFile);
	  else
	     filename.printf("%s.Empirical.Kinship",(const char *) PreMeta::outputFile);
       }
       else
       {
	  filename.printf("Empirical.Kinship");
       }
       file = fopen(filename, "wt");
       for(int p=0;p<total_n;p++)
       {
	  fprintf(file,"%s ",ped[genotypedSamplePED[p]].pid.c_str());
       }
       fprintf(file,"\n");

       for(int r=0;r<allPairs.rows;r++)
       {
	  for(int c=0;c<allPairs.cols;c++)
	  {
	     fprintf(file,"%g ",allPairs[r][c]);
	  }
	  fprintf(file,"\n");
       }
       fclose(file);
    }
}

void KinshipEmp::SetupVCF(Pedigree & ped, IntArray * pPheno,IntArray & genotypedSampleVCF,FILE * log)
{
   printf("Calculating empirical kinship matrix...\n");
   fprintf(log,"Calculating empirical kinship matrix...\n");
   VcfFileReader reader;
   VcfHeader header;
   VcfRecord record;
   reader.open(PreMeta::vcfInput,header);

   int N=0;
   int total_n = genotypedSampleVCF.Length();
   allPairs.Dimension(total_n,total_n);
   allPairs.Zero();

   Vector genotype;
   genotype.Dimension(total_n);

   String chr;
   int pos;

   if(PreMeta::dosage)
   {
      while(reader.readRecord(record))
      {
	 chr = record.getChromStr();
	 pos = record.get1BasedPosition();

	 if(chr==PreMeta::xLabel && pos>=PreMeta::Xstart && pos<=PreMeta::Xend)
	    continue;

	 double mean = 0.0, mean_sqr= 0.0,maf=0.0;
	 int n_=0, nmiss=0;
	 VcfRecordGenotype & genoInfo = record.getGenotypeInfo();
	 for (int p = 0; p < total_n; p++)
	 {
	    int s = genotypedSampleVCF[p];
	    //fill in genotype vector for this sample
	    const std::string * geno = genoInfo.getString("DS",s);
	    if(*geno == ".")
	    {
	       genotype[p] = _NAN_;
	       nmiss++;
	    }
	    else
	    {
	       genotype[p] = atof((*geno).c_str());
	       mean += genotype[p];
	       n_++;
	       mean_sqr += genotype[p]*genotype[p];
	    }
	 }
	 maf = mean/(2.0*n_);
	 if(nmiss>total_n*miss || maf<q || 1.0-maf<q)
	    continue;

	 N++;
	 mean = maf*2.0;
	 double variance = mean_sqr/n_ - mean*mean;
	 //standardize genotype vector
	 for(int i=0;i<genotype.Length();i++)
	 {
	    if(genotype[i]==_NAN_)
	       //make sure if genotype is missing, then it is not contributing
	       genotype[i]=0.0;
	    else
	       genotype[i] -= mean;
	 }
	 //add up to kinship
	 for(int r=0;r<total_n;r++)
	 {
	    for(int c=r;c<total_n;c++)
	    {
	       allPairs[r][c] += genotype[r]*genotype[c]/variance;
	    }
	 }
      }
      for(int i=0;i<total_n;i++)
	 for(int j=i+1;j<total_n;j++)
	    allPairs[j][i] = allPairs[i][j];

      allPairs.Multiply(1.0/N);
   }
   else
   {
      IntArray counts;
      counts.Dimension(2);
      while(reader.readRecord(record))
      {
	 chr = record.getChromStr();
	 pos = record.get1BasedPosition();

	 if(chr==PreMeta::xLabel && pos>=PreMeta::Xstart && pos<=PreMeta::Xend)
	    continue;

	 double mean = 0.0, mean_sqr= 0.0,maf=0.0;
	 int n_=0, nmiss=0;
	 counts.Zero();
	 genotype.Zero();

	 bool skipSNP=false;
	 for(int s=0;s<total_n;s++)
	 {
	    bool skip=false;
	    int i= genotypedSampleVCF[s];
	    int numGTs = record.getNumGTs(i);
	    for(int j = 0; j < numGTs; j++)
	    {
	       int a = record.getGT(i,j);
	       if(a==VcfGenotypeSample::MISSING_GT)
	       {
		  nmiss++;
		  if(nmiss>total_n*miss)
		  {
		     skipSNP=true;
		     break;
		  }
		  skip =true;
		  genotype[s]=_NAN_;
		  break;
	       }
	       //if marker is not biallelic, skip this marker
	       if(a>1)
	       {
		  skipSNP=true;
		  break;
	       }
	       genotype[s] += a;
	       counts[a]++;
	    }
	    if(skipSNP)
	       break;
	    if(skip)
	       continue;
	    mean+=genotype[s];
	    n_++;
	    mean_sqr += genotype[s]*genotype[s];
	 }

	 if(skipSNP)
	    continue;

	 maf = (double)counts[1]/(counts[0]+counts[1]);

	 if(nmiss > total_n*miss || maf<q || 1.0-maf<q)
	    continue;

	 N++;
	 mean /= n_;
	 double variance = mean_sqr/n_ - mean*mean;

	 for(int i=0;i<genotype.Length();i++)
	 {
	    if(genotype[i]==_NAN_)
	       genotype[i]=0.0;
	    else
	       genotype[i] -= mean;
	 }

	 for(int r=0;r<total_n;r++)
	 {
	    for(int c=r;c<total_n;c++)
	    {
	       allPairs[r][c] += genotype[r]*genotype[c]/variance;
	    }
	 }
      }

      for(int i=0;i<total_n;i++)
	 for(int j=i+1;j<total_n;j++)
	    allPairs[j][i] = allPairs[i][j];

      allPairs.Multiply(1.0/N);
   }

   if(N==0)
      error("ERROR! No qualified marker was included in calculating empirical kinsihp matrix.\n");


   printf("%d markers were included to calculate empirical kinship.\n",N);
   fprintf(log,"%d markers were included to calculate empirical kinship.\n",N);

   if(OutputKin::outputKin)
   {
      printf("Writing kinship matrix ...\n");
      fprintf(log,"Writing kinship matrix ...\n");
      FILE * file = NULL;
      String filename;
      if(PreMeta::outputFile != "")
      {
	 if(PreMeta::outputFile.FastFindLastChar('.')!=-1 || PreMeta::outputFile.FastFindLastChar('/')!=-1)
	    filename.printf("%sEmpirical.Kinship",(const char *) PreMeta::outputFile);
	 else
	    filename.printf("%s.Empirical.Kinship",(const char *) PreMeta::outputFile);
      }
      else
	 filename.printf("Empirical.Kinship");
      file = fopen(filename, "wt");
      for (int i = 0; i < total_n; i++)
      {
	 const char * sample = header.getSampleName(genotypedSampleVCF[i]);
	 fprintf(file,"%s ",sample);
      }
      fprintf(file,"\n");
      for(int r=0;r<allPairs.rows;r++)
      {
	 for(int c=0;c<allPairs.cols;c++)
	 {
	    fprintf(file,"%g ",allPairs[r][c]);
	 }
	 fprintf(file,"\n");
      }
      fclose(file);
   }
}
